/*      thread_qsort.cxx -- A test driver of multithreaded quick sort.

        CSC402, Fall, 2009, Dr. Dale Parson.
*/

// Usage: thread_qsort NUMBER_OF_INTEGERS THREADING_THRESHOLD RANDOM_SEED
static const char *usage =
    "usage: thread_qsort NUMBER_OF_INTEGERS THREADING_THRESHOLD RANDOM_SEED" ;

// THE FOLLOWING EXIT CODES CONSTITUTE THE RETURN VALUES FROM MAIN:
static const int    SUCCESS     =       0 ; /* successful program execution */
static const int    PARMERR     =       1 ; /* invalid command line values */
static const int    THREADERR     =     2 ; /* problem with threading */

#include <iostream>
#include <stdlib.h>
#include "StringToInteger.h"
#include <sys/types.h>
#include <time.h>

using namespace std ;

static void quicksort(int iarray[], int left, int right);
static void printarray(int iarray[], int numelems) ;

// threadingThreshold is global to file to avoid parameter pass overhead
static int threadingThreshold = 0 ;

/*
        Function:           main

            main is a test driver that exercises implementations of
            a quick sort and binary search.

        Parameters:

            See usage above.

        Return value:       See EXIT CODES above.

*/
int
main(const int argc, char *argv[]) {
    int numberOfIntegers = 0, randomSeed = 0 ;
    bool validFlag = true ;
    if (argc != 4) {
        cerr << usage << endl ;
        exit(PARMERR);
    }
    numberOfIntegers = StringToInteger(argv[1], validFlag);
    if ((! validFlag) || numberOfIntegers < 1) {
        cerr << "Invalid NUMBER_OF_INTEGERS: " << argv[1] << endl ;
        cerr << usage << endl ;
        exit(PARMERR);
    }
    threadingThreshold = StringToInteger(argv[2], validFlag);
    if ((! validFlag) || numberOfIntegers < 0) {
        cerr << "Invalid THREADING_THRESHOLD: " << argv[2] << endl ;
        cerr << usage << endl ;
        exit(PARMERR);
    }
    randomSeed = StringToInteger(argv[3], validFlag);
    if (! validFlag) {
        cerr << "Invalid RANDOM_SEED: " << argv[1] << endl ;
        cerr << usage << endl ;
        exit(PARMERR);
    }
    srand(randomSeed);
    int *intarray = new int [ numberOfIntegers ];
    for (int i = 0 ; i < numberOfIntegers ; i++) {
        intarray[i] = rand();
    }
    // CLOCK MAY MEASURE THE SAUM OF THE CPUs. Use wall clock time.
    // int before = clock();
    int before = time(0);
    quicksort(intarray, 0, numberOfIntegers-1);
    // int after = clock();
    int after = time(0);
    cerr << "SORT TIME IN uSECS = " << (after - before) << endl ;
    // ^^^ left and right are first and last indices
    // printarray(intarray, numberOfIntegers);
    return(SUCCESS);
}

// If there are at least 3 elements, pick a value that is >= one
// of them and <= the other as the pivot. This decreases the probability
// of hitting a degenerate case where all or almost all elements partition
// to one side of the pivot.
// The return value is the pivot, also swapped into iarray[left].
static int pickpivot(int iarray[], int left, int right) {
    int result ;
    int count = right - left + 1 ;
    int mid = (left + right) / 2 ;
    if (count < 3) {
        result = iarray[left];  // [1] element or 50/50 odds for [2]
    } else if ((iarray[mid] <= iarray[left] && iarray[left] <= iarray[right])
            || (iarray[right] <= iarray[left] && iarray[left] <= iarray[mid])){
        result = iarray[left];
    } else if ((iarray[left] <= iarray[mid] && iarray[mid] <= iarray[right])
            || (iarray[right] <= iarray[mid] && iarray[mid] <= iarray[left])){
        result = iarray[mid] ;
        iarray[mid] = iarray[left] ;
        iarray[left] = result ;
    } else if ((iarray[mid] <= iarray[right] && iarray[right] <= iarray[left])
            || (iarray[left] <= iarray[right] && iarray[right] <= iarray[mid])){
        result = iarray[right] ;
        iarray[right] = iarray[left] ;
        iarray[left] = result ;
    }
    return result ;
}

// partition is a helper function for quicksort().
// partition partitions an array into a left "half" and a right "half,"
// where all values in the left "half" or <= some "pivot" value, and
// all values in the  right "half" are > that "pivot" value.
// partition may not divide the array exactly in half; left and right
// "part" would be a better term.
// RETURN index of the element around which the array is partitioned.
static int partition(int iarray[], int left, int right) {
    int pivot = pickpivot(iarray, left, right);
    int low = left + 1 ;        // scan up looking for a too-high value
    int high = right ;          // scan down looking for a too-low value

    while (low < high) {      // scan from both sides of array
        while (low <= high && iarray[low] <= pivot) { // in correct side
            low++;
        }
        while (low <= high && iarray[high] > pivot) { // in correct side
            high--;
        }
        // If there are two elements on the wrong side, swap them.
        if (low < high) {
            int temp = iarray[high];
            iarray[high] = iarray[low];
            iarray[low] = temp ;
        }
    }
    // We still need to place the pivot in the correct spot.
    while (high > left && iarray[high] >= pivot) {
        high--;
    }
    // swap pivot with a lower value if there is one
    if (pivot > iarray[high]) {
        iarray[left] = iarray[high];
        iarray[high] = pivot ;
        return high;
    } else {
        return left ;
    }
}

struct thread_param {     // always allocate on a non-returned stack frame!
    int * iarray ;      // subarray to be sorted
    int left, right ;   // it's left and right indices
} ;

static void *threadstart(void *voidarg) {
    thread_param *arg = (thread_param *) voidarg ;
    // sort partition in this thread
    quicksort(arg->iarray, arg->left, arg->right);
    return 0 ;  // thread terminates
}

// Implement the quick sort algorithm.
static void quicksort(int iarray[], int left, int right) {
    int count = right - left + 1 ;
    if (left < right) {
        // Partition array so all values <= iarray[pivotlocation] are left
        // of pivotlocation, and all values > pivotlocation are right of it.
        int pivotlocation = partition(iarray, left, right);
        // Sort the left partition.
        if (count >= threadingThreshold) {
            // Create another thread to sort the left partition.
            // Package up arguments for the thread:
            pthread_t child ;
            thread_param thread_arg ;
            thread_arg.iarray = iarray ;
            thread_arg.left = left ;
            thread_arg.right = pivotlocation-1 ;
            if (pthread_create(&child, 0, threadstart, &thread_arg)) {
                cerr << "ERROR STARTING A THREAD." << endl ;
                exit(THREADERR);
            }
            // Sort the right partition while the child thread sorts the left.
            quicksort(iarray, pivotlocation+1, right);
            // Wait for the child to complete the left and terminate.
            if (pthread_join(child, 0)) {
                cerr << "ERROR JOINING A THREAD." << endl ;
                exit(THREADERR);
            }
        } else {
            // Sort the left partition.
            quicksort(iarray, left, pivotlocation-1);
            // Sort the right partition.
            quicksort(iarray, pivotlocation+1, right);
        }
    }
}

static void printarray(int iarray[], int numelems) {
    for (int i = 0 ; i < numelems ; i++) {
        cout << iarray[i] << endl  ;
    }
}
